# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.

# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import asyncio
import aiohttp
import importlib
import logging
import os
import hashlib
import secrets

from aiohttp import web

import config
import http_client
import paths
import web_interface.page_former
import web_interface.printers_html
import user_login
import version


PATHS_WITH_FILE_UPLOAD = {'accept_update_from_file' : paths.UPDATE_FILE_PATH,\
                          'upload_plugin' :  paths.PLUGIN_INSTALL_FILE_PATH,\
                          'upload_to_storage' : None}
EXECUTOR_TIMEOUT = 5.8 # 0.2s less than WebInterface.SHUTDOWN_TIMEOUT
UPLOAD_CHUNK_SIZE = 1024*256 # greatly affects upload performance
AUTH_ERROR_SLEEP = 1


def check_auth(request):
    try:
        host = request.host.split(':')[0]
    except:
        host = "127.0.0.1"
    if request.remote != host:
        if request.secure:
            cookie = request.cookies.get(web_interface.page_former.PageFormer.COOKIE_NAME_HTTPS)
        else:
            cookie = request.cookies.get(web_interface.page_former.PageFormer.COOKIE_NAME)
        if not cookie or cookie not in request.app['cookie_tokens']:
            return "Error: not authenticated"


async def reload_page_former(request):
    if request.remote == request.host.split(':')[0]:
        logging.getLogger('app.reload_page_former').info("Reloading the PageFormer")
        importlib.reload(web_interface.page_former)
        raise web.HTTPNoContent()


async def generic(request):
    logger = logging.getLogger('app.generic_web_handler')
    pf_args = []
    if request.method == 'GET':
        logger.debug("GET %s" % request.path)
        method_name = 'do_GET'
    elif request.method == 'POST':
        logger.debug("POST %s" % request.path)
        method_name = 'do_POST'
        post_data = await request.post()
        pf_args.append(post_data)
    else:
        raise web.HTTPForbidden()
    return await form_page_and_respond(request, method_name, pf_args)


async def multipart_upload_handler(request):
    logger = logging.getLogger('app.multipart_upload_handler')
    error = check_auth(request)
    file_path = PATHS_WITH_FILE_UPLOAD.get(request.path.strip("/"))
    logger.info("Upload path: %s to %s" % (request.path, file_path)) 
    if not error:
        try:
            reader = await request.multipart()
            fields_left = 128 # max number of fields
            while True:
                if request.app['3dp_app'].stop_flag:
                    raise web.HTTPFound("/")
                field = await reader.next()
                if field and field.name == 'file':
                    break
                if fields_left:
                    fields_left -= 1
                else:
                    raise RuntimeError('No upload got not file field')
        except Exception as e:
            error = "Error while searching for upload field: " + str(e)
            logger.error(error)
        else:
            if not file_path:
                filename = field.filename
                if filename:
                    file_path = os.path.join(paths.STORAGE_FOLDER, filename)
                else:
                    error = "Please select a file"
    if not error:
        size = 0
        try:
            with open(file_path, 'wb') as f:
                while True:
                    if request.app['3dp_app'].stop_flag:
                        break
                    chunk = await field.read_chunk(UPLOAD_CHUNK_SIZE) 
                    if not chunk:
                        break
                    size += len(chunk)
                    file_write_future = asyncio.get_event_loop().run_in_executor(
                            request.app['threads_pool'],
                            lambda f, chunk: f.write(chunk),
                            f,
                            chunk)
                    await asyncio.wait_for(file_write_future, EXECUTOR_TIMEOUT)
            if request.app['3dp_app'].stop_flag:
                os.remove(file_path)
                raise web.HTTPFound("/")
        except (OSError, IOError) as e:
            error = "Error saving upload: " + str(e)
            logger.error(error)
        else:
            logger.info('Received upload size: %dB' % size)
    pf_args = [{"multipart_error" : error}]
    return await form_page_and_respond(request, 'do_POST', pf_args)


async def form_page_and_respond(request, method_name, pf_args=[]):
    logger = logging.getLogger('app.generic_web_handler')
    pf = web_interface.page_former.PageFormer(request.app['3dp_app'],
                     request.app['3dp_content'],
                     request.host,
                     request.method,
                     request.raw_path,
                     request.secure,
                     request.remote,
                     request.cookies,
                     request.app['cookie_tokens'],
                     request.app['wizard_stage'],
                     False)
    try:
        executor_future = asyncio.get_event_loop().run_in_executor(request.app['threads_pool'], getattr(pf, method_name), *pf_args)
        await asyncio.wait_for(executor_future, EXECUTOR_TIMEOUT)
    except asyncio.TimeoutError:
        logger.warning(f'Executor timeout when handling {request.raw_path}')
        pf.status = 504
    except Exception as e:
        logger.error(e) # check why some error are not logged by log.log_exception
    else:
        if pf.wizard_former.current_stage != request.app['wizard_stage']:
            logger.info("Wizard stage: " + str(pf.wizard_former.current_stage))
            request.app['wizard_stage'] = pf.wizard_former.current_stage
        request.app['wizard_enabled'] = pf.wizard_enabled
        if pf.new_cookie_token:
            if pf.new_cookie_token not in request.app['cookie_tokens']:
                request.app['cookie_tokens'].append(pf.new_cookie_token)
                request.app.save_cookie_tokens()
    return web.Response(status=pf.status, headers=pf.headers, body=pf.body)


async def printers_html(request):
    return web.Response(body=web_interface.printers_html.get_printers_html(request))


async def http_warning_html(request):
    body = ''
    if not request.secure:
        https_port = config.get_settings()['web_interface']['port_https']
        if https_port:
            http_port = config.get_settings()['web_interface']['port']
            host = request.host.strip(":" + str(http_port))
            if http_port != 443:
                host += ":" + str(https_port)
            body = '<div id="http_secure_message">For security reasons we strongly recommend to use HTTPS url to access the cloud client interface.' \
                    f' <a href="https://{host}/">Click here to switch to HTTPS</a>.</div>'
    return web.Response(body=body)
